xquery version "1.0-ml" ;

(: cluster the Shakespeare speeches, disregarding the speaker, 
   and show the results in an html table :)

declare namespace db="http://marklogic.com/xdmp/database" ;
declare namespace cl="cts:cluster" ;
declare namespace dt="cts:distinctive-terms" ;

(: generally we want to cluster the top N results, where N is 
   around 100 to 1,000 (smaller numbers for best performance).
   all speeches = 31,029; 
   speeches that contain "love" = 1,864; 
   "war" = 359; "joy" = 201; 
   "beast" = 94; 
   "aunt"=24
:)
let $search-term := xdmp:get-request-field("search-term", "global")
let $word:="adaptation"
let $max-terms := xdmp:get-request-field("max-terms", "100")   
let $use-db-config := 
  xdmp:get-request-field("use-db-config", "false")   

let $options-node :=
  <options xmlns="cts:cluster" xmlns:db="http://marklogic.com/xdmp/database">
    <hierarchical-levels>0</hierarchical-levels>
    <algorithm>lsi</algorithm>
    <overlapping>false</overlapping>
    <label-max-terms>10</label-max-terms>
   
    <max-clusters>30</max-clusters>
    <label-ignore-words>of the on in at a an for from by and</label-ignore-words>
    <db:stemmed-searches>advanced</db:stemmed-searches>
    <db:fast-phrase-searches>true</db:fast-phrase-searches>
    <db:fast-element-word-searches>true</db:fast-element-word-searches>
    <db:fast-element-phrase-searches>true</db:fast-element-phrase-searches>
  </options>

(: build the page :)
let $page :=
<html>
<head><title>Example - clustering hypothesis</title></head>
<body>
<table border="1" cellpadding="1" cellspacing="1">
  <tr>
    <th>Label</th>
    <th>Count</th>
    <th>Content</th>
  </tr>
  {
let $things-to-cluster := 
  cts:search(
     (: specify the directory in which you have loaded the plays :)
       collection("hypothesis")//h,
       cts:and-query(()) 
        )
(: iterate through the cts:cluster results node :)
for $cluster in 
    cts:cluster( $things-to-cluster, $options-node )/cts:cluster
return 
    <tr>
     <td>{ fn:data( $cluster/@label ) }</td>
     <td>{ fn:data( $cluster/@count ) }</td>
     <td>
       <table>{
 for $clustered-node-ref in fn:data( $cluster/@nodes )
 return
     <tr><td>{ fn:string( 
       $things-to-cluster[$clustered-node-ref] ) 
     }</td></tr>
       }</table>
      </td>
     </tr>}
</table>
</body>
</html>

return ( xdmp:set-response-content-type("text/html"), 
 $page, xdmp:elapsed-time() )